import os
import requests
import numpy as np
from PIL import Image
from scipy.optimize import minimize
from skimage.feature import register_translation
from scipy.ndimage import shift

# from . import costLib as ctl
# from . import mtfLib as mtl
# from . import imageLib as iml
# from . import costDerivativesLib as cdl
import costLib as ctl
import mtfLib as mtl
import imageLib as iml
import costDerivativesLib as cdl


from matplotlib import pyplot as plt

def get_mapfost_path():
    return os.path.dirname(__file__)


def est_aberr(test_ims, test_aberrs, pix_size_um=0.08, num_aperture=0.004, stig_rot_deg=0,
              stig_scale=1, use_bessel=False, crop_ref=(0,0), crop_size=(512,512),
              test_ims_aligned=0, get_hessian=False, get_uncertainty=False, do_subpixel_shift=False):
    test_ims_cropped = [
        np.array(Image.fromarray(m).crop((crop_ref[0], crop_ref[1], crop_ref[0] + crop_size[0], crop_ref[1] + crop_size[1]))) for m
        in test_ims]

    if not test_ims_aligned:
        shift_in_meas = [iml.get_shift_vec(test_ims_cropped[0], test_ims_cropped[1]), [0, 0]]
    else:
        shift_in_meas = [[0, 0], [0, 0]]

    for ish in [0,1]:
        if shift_in_meas[0][ish]<0:
            shift_in_meas[1][ish] = shift_in_meas[0][ish]*-1
            shift_in_meas[0][ish] = 0
    # all_crop_refS = [[crop_ref[0] + shift_in_meas[k][0],
    #                   crop_ref[1] + shift_in_meas[k][1],
    #                   crop_ref[0] + crop_size[0] + shift_in_meas[k][0],
    #                   crop_ref[1] + crop_size[1] + shift_in_meas[k][1]] for k in [0, 1]]

    #TODO: add the crop valid check
    cropping_valid = True

    if cropping_valid:
        test_ims_cr_al = [np.array(Image.fromarray(m).crop((crop_ref[0] + shift_in_meas[k][0], crop_ref[1] + shift_in_meas[k][1],
                                           crop_ref[0] + crop_size[0] + shift_in_meas[k][0],
                                           crop_ref[1] + crop_size[1]
                                           + shift_in_meas[k][1]))) for k, m in enumerate(test_ims)]

        if do_subpixel_shift:
            shifted, error, diffphase = register_translation(test_ims_cr_al[0], test_ims_cr_al[1], 100)
            test_ims_cr_al[1] = shift(test_ims_cr_al[1], shift=(shifted[0], shifted[1]), mode='constant')

        test_ims_cr_al_pr = [iml.removeMeanAndMeanSlope(im) for im in test_ims_cr_al]
        test_ims_cr_al_pr_fft = [np.fft.fftshift(np.fft.fft2(im)) for im in test_ims_cr_al_pr]
        test_ims_cr_al_pr_fft_lp = [iml.low_pass_filter(m) for m in test_ims_cr_al_pr_fft]

        ps_width, ps_height = test_ims_cr_al_pr_fft_lp[0].shape
        kspace_xy = mtl.get_kspace_xy(ps_width, pix_size_um, stig_rot_deg)

        res = minimize(ctl.cost_mapfost, [0, 0, 0],
                       args=(test_ims_cr_al_pr_fft_lp, test_aberrs, num_aperture,stig_scale, use_bessel, kspace_xy),
                       method="L-BFGS-B")
        res.x = np.round(res.x,2)

        if get_hessian:
            hess_field = cdl.get_hessian_matrix(test_ims_cr_al_pr_fft_lp, np.add(test_aberrs, res.x), num_aperture, pix_size_um)
            res.hess_field = hess_field
            if get_uncertainty:
                res.uncertainty = 1/np.sum(np.diagonal(hess_field**2))

        return res


def test_installation():

    im1 = np.array(Image.open(requests.get('https://gitlab.com/rangolisaxena90/mapfost/-/raw/main/src/mapfost/example/test_im1.png', stream=True).raw))
    im2 = np.array(Image.open(requests.get('https://gitlab.com/rangolisaxena90/mapfost/-/raw/main/src/mapfost/example/test_im2.png', stream=True).raw))
    res = est_aberr([im1, im2], [[-4,0,0],[4,0,0]], get_hessian=True)
    if res.x[0] == -1.11 and res.x[1] == -1.31 and res.x[2] == 0.59:
        print("...Installation Test Successful...")
        print(res)
    else:
        print(".. check if the result is close to the one given in the docs...")
        print(res)




import numpy as np

def firstDerivAx(edge_len,z,ax):

    wngrd = np.array(
        np.meshgrid(range(int(-edge_len / 2), int(edge_len / 2)),
                    range(int(-edge_len / 2), int(edge_len / 2)))[::-1],
        dtype=np.float)

    wtd =  np.add(
            np.multiply(
                np.multiply(
                    wngrd[0], wngrd[0]),
                2 * ax + 2 * z),
            np.multiply(
                np.multiply(
                    wngrd[1], wngrd[1]),
                2 * ax - 2 * z))

    return wtd


def firstDerivAy(edge_len, z, ay):

    wngrd = np.array(
        np.meshgrid(range(int(-edge_len / 2), int(edge_len / 2)),
                    range(int(-edge_len / 2), int(edge_len / 2)))[::-1],
        dtype=np.float)

    wtd = np.add(
        np.add(
            np.multiply(
                np.multiply(
                    wngrd[0], wngrd[0]),
                2 * ay),
            np.multiply(
                np.multiply(
                    wngrd[1], wngrd[1]),
                2 * ay)),
        np.multiply(
            np.multiply(
                wngrd[0], wngrd[1]), -4 * z))

    return wtd


def firstDerivZ(edge_len, z, ax, ay):

    wngrd =np.array(
        np.meshgrid(range(int(-edge_len / 2), int(edge_len / 2)),
                    range(int(-edge_len / 2), int(edge_len / 2)))[::-1],
        dtype=np.float)

    wtd = np.add(
        np.add(
            np.multiply(
                np.multiply(
                    wngrd[0], wngrd[0]),
                    2 * ax + 2 * z),
            np.multiply(
                np.multiply(
                    wngrd[1], wngrd[1]),
                     -2 * ax + 2 * z)),
        np.multiply(
            np.multiply(
                wngrd[0], wngrd[1]), -4 * ay))

    return wtd


def secondDerivAxOrAyOrZ(w,pxsz):

    wngrd = np.meshgrid(range(int(-w/2), int(w/2)), range(int(-w/2), int(w/2))[::-1])
    wnc = 2*np.pi/int(w*pxsz)
    nn = 2*((wngrd[0] * wnc)**2 + (wngrd[1] * wnc)**2)
    return nn

def secondDerivAxZ(w,pxsz):

    wngrd = np.meshgrid(range(int(-w/2), int(w/2)), range(int(-w/2), int(w/2))[::-1])
    wnc = 2*np.pi/int(w*pxsz)
    nn = 2*((wngrd[0] * wnc)**2 - (wngrd[1] * wnc)**2)
    return nn


def secondDerivAyZ(w,pxsz):

    wngrd = np.meshgrid(range(int(-w/2), int(w/2)), range(int(-w/2), int(w/2))[::-1])
    wnc = 2*np.pi/int(w*pxsz)
    pp = -4*wngrd[0]*wngrd[1]*wnc**2
    return pp


def secondDerivAyAx(w,pxsz):

    wngrd = np.meshgrid(range(int(-w/2), int(w/2)), range(int(-w/2), int(w/2))[::-1])
    pp = wngrd[0]*wngrd[1]*0

    return pp


def getd1MTF(mtf, z, ax, ay, num_aperture, pxsz):

    edge_len = mtf.shape[0]
    wnCoeff = 2 * np.pi / (int(edge_len) * pxsz)
    mtfFirstDiv = [firstDerivZ(edge_len,z,ax,ay), firstDerivAx(edge_len,z,ax),  firstDerivAy(edge_len,z,ay)]
    d1mtf = [np.multiply(mtf,-(1/8)*wnCoeff*num_aperture**2*imtf) for imtf in mtfFirstDiv]

    return d1mtf


def getd2MTF(mtf, z, ax, ay, num_aperture, pxsz):
    edge_len = mtf.shape[1]
    mtfFirstDiv = getd1MTF(mtf, z, ax, ay, num_aperture, pxsz)
    mtfSecondDiv = [secondDerivAxOrAyOrZ(edge_len, pxsz), secondDerivAxZ(edge_len,pxsz),secondDerivAyZ(edge_len,pxsz),
                    secondDerivAxZ(edge_len,pxsz), secondDerivAxOrAyOrZ(edge_len, pxsz), secondDerivAyAx(edge_len,pxsz),
                    secondDerivAyZ(edge_len,pxsz),secondDerivAyAx(edge_len,pxsz),  secondDerivAxOrAyOrZ(edge_len, pxsz)]

    iv, jv = np.indices((3,3))

    iv = iv.reshape(-1)
    jv = jv.reshape(-1)

    res = [None]*9
    for inx in range(0,9):

        A = (-1/8)*num_aperture**2

        i = iv[inx]
        j = jv[inx]
        ij = 3*i +j
        fin1 = A**2*np.multiply(mtf, np.multiply(mtfFirstDiv[i], mtfFirstDiv[j]))
        fin2 = A*np.multiply(mtf, mtfSecondDiv[ij])

        fin = np.add(fin1,fin2)

        res[inx] = fin

    return res


def getMTF(edge_len, z, ax, ay, num_aperture, pxsz):

    wnc = 2 * np.pi/ (int(edge_len)*pxsz)
    wngrd = wnc * np.array(
                        np.meshgrid(range(int(-edge_len / 2), int(edge_len / 2)),
                                    range(int(-edge_len / 2), int(edge_len / 2)))[::-1],
                        dtype = np.float)

    wtd = np.add(
            np.add(
                np.multiply(
                        np.multiply(
                                wngrd[0], wngrd[0]),
                                z ** 2 + ax ** 2 +  ay ** 2 + 2 * ax * z),
                np.multiply(
                        np.multiply(
                                wngrd[1], wngrd[1]),
                                z ** 2 + ax ** 2 +  ay ** 2 - 2 * ax * z)),
                np.multiply(
                    np.multiply(
                        wngrd[0], wngrd[1]), -4 * ay * z))

    mtf = np.exp(np.multiply((-1 / 8) * (num_aperture ** 2), wtd))

    return mtf


def get_jacobian_vec(test_ims, total_aberrs, num_aperture, pix_size_um, fadeFac, mtfSqS=None, d1mtfSqS = None):

    edge_len = np.array(test_ims).shape[1]
    if mtfSqS is None:
        mtf1 = getMTF(edge_len, *total_aberrs[0], num_aperture, pix_size_um) * fadeFac[0]
        mtf2 = getMTF(edge_len, *total_aberrs[1], num_aperture, pix_size_um) * fadeFac[1]

    if d1mtfSqS is None:
        d1mtf1 = getd1MTF(mtf1, *total_aberrs[0], num_aperture, pix_size_um)
        d1mtf2 = getd1MTF(mtf2, *total_aberrs[1], num_aperture, pix_size_um)

    jacobian_vec = [None] * 3
    x1, x2 = test_ims
    for i in range(0, 3):

        aa = np.subtract(np.multiply(x1, mtf2**2), np.multiply(x2, mtf1**2))

        bb = np.subtract(np.multiply(x1, np.multiply(mtf2,d1mtf2[i])),
                         np.multiply(x2, np.multiply(mtf1,d1mtf1[i])))
        cc = np.add(mtf1**4, mtf2**4)
        dd = np.add(np.multiply(mtf1**3, d1mtf1[i]), np.multiply(mtf2**3, d1mtf2[i]))

        add1 = 4*np.divide(np.multiply(aa, bb), cc)
        add2 = -4*np.divide(np.multiply(aa**2, dd), cc**2)
        fin = np.add(add1,add2)
        jacobian_vec[i] = fin

    jacobian_vec = np.array(jacobian_vec).T.reshape(edge_len * edge_len, 3)

    return np.nansum(jacobian_vec, axis=0)

def get_hessian_matrix(test_ims, total_aberrs, num_aperture, pix_size_um):

    edge_len = np.array(test_ims).shape[1]
    mtf1= getMTF(edge_len, *total_aberrs[0], num_aperture, pix_size_um)
    mtf2 = getMTF(edge_len, *total_aberrs[1], num_aperture, pix_size_um)

    d1mtf1 = getd1MTF(mtf1, *total_aberrs[0], num_aperture, pix_size_um)
    d1mtf2 = getd1MTF(mtf1, *total_aberrs[1], num_aperture, pix_size_um)

    d2mtf1 = getd2MTF(mtf1, *total_aberrs[0], num_aperture, pix_size_um)
    d2mtf2 = getd2MTF(mtf1, *total_aberrs[1], num_aperture, pix_size_um)

    hessian_field = [[None]] * 9
    iv, jv = np.indices((3, 3))

    iv = iv.reshape(-1)
    jv = jv.reshape(-1)

    x1 = np.real(test_ims[0])
    y1 = np.imag(test_ims[0])

    x2 = np.real(test_ims[1])
    y2 = np.imag(test_ims[1])

    for inx in range(0, 9):

        i = iv[inx]
        j = jv[inx]

        ij = i * 3 + j
        aa = x1 * d1mtf2[j] - x2 * d1mtf1[j]
        bb = x1 * d1mtf2[i] - x2 * d1mtf1[i]
        cc = y1 * d1mtf2[j] - y2 * d1mtf1[j]
        dd = y1 * d1mtf2[i] - y2 * d1mtf1[i]

        oo = x1 * mtf2 - x2 * mtf1
        nn = x1 * d2mtf2[ij] - x2 * d2mtf1[ij]

        ll = y1 * mtf2 - y2 * mtf1
        mm = y1 * d2mtf2[ij] - y2 * d2mtf1[ij]

        ee = mtf1 * d1mtf1[i] + mtf2 * d1mtf2[i]
        kk = mtf1 * d1mtf1[j] + mtf2 * d1mtf2[j]
        ff = mtf1 * d2mtf1[ij] + mtf2 * d2mtf2[ij] + d1mtf1[i] * d1mtf1[j] + d1mtf2[i] * d1mtf2[j]
        gg = mtf1**2 + mtf2**2

        A = np.add(oo**2, ll**2)
        Aij = 2 *np.sum([np.multiply(aa,bb), np.multiply(cc,dd), np.multiply(oo,nn), np.multiply(ll,mm)], axis=0)
        Ai = 2 *np.sum([np.multiply(oo,bb), np.multiply(ll,dd)], axis=0)
        Aj = 2 *np.sum([np.multiply(oo,aa), np.multiply(ll,cc)], axis=0)

        Bi = 2 * ee
        Bj = 2 * kk
        Bij = 2 * ff

        B = gg
        fin1 = np.multiply(Aij, np.multiply(B, B))
        fin2 = -1 * np.multiply(Ai, np.multiply(Bj, B))
        fin3 = -1 * np.multiply(Aj, np.multiply(Bi, B))
        fin4 = 2 * np.multiply(A, np.multiply(Bi, Bj))
        fin5 = -1 * np.multiply(A, np.multiply(Bij, B))
        fin = np.divide(np.add(fin1, np.add(fin2, np.add(fin3, np.add(fin4, fin5)))), B ** 3)
        hessian_field[inx] = fin
    hessian = np.nansum(np.array(hessian_field).T.reshape(edge_len*edge_len, 3, 3), axis=0)

    return hessian




import numpy as np
from scipy import special

def spaceRotate(meshx, meshy, rot):
    """
    rotates the spatial mesh for x and spatial mesh for y by the angle rot.
    Args:
        meshx: 2D array spatial mesh for x
        meshy: 2D array spatial mesh for y
        rot: rotation angle in degrees

    Returns:
        rotated spatial mesh
    """
    meshxtr = np.subtract(meshx * np.cos(np.radians(rot)), meshy * np.sin(np.radians(rot)))
    meshytr = np.add(meshx * np.sin(np.radians(rot)), meshy * np.cos(np.radians(rot)))
    return [meshxtr, meshytr]

def get_kspace_xy(edgeLen, pixSzUM, astigRot=0, pow_spec_crop_factor=0.5):
    """
    Returns the wave vectors kx and ky of size edgeLen and rotated by the angle astigRot
    Args:
        edgeLen: the edge length of the wave vector grid
        pixSzUM: pixel size in um
        astigRot: the rotation in degrees between the image plane and the stigmation plane
        pow_spec_crop_factor: ratio of the filtered edge length to the inital edge length
         of the test images power spectrum. Refer mapfost.imageLib.lowpass_pow_spec()

    Returns:
        2D array of kx and ky wave vectors in a list of length 2 [kx, ky]. rotated by astigRot
    """

    kNyquist = 2*np.pi*pow_spec_crop_factor/pixSzUM

    kXarr = np.linspace(-kNyquist, kNyquist, edgeLen, endpoint=False)
    kYarr = np.linspace(-kNyquist, kNyquist, edgeLen, endpoint=False)[::-1]
    kXY = np.meshgrid(kXarr, kYarr)

    kXYrotated = spaceRotate(kXY[0], kXY[1],astigRot)

    return kXYrotated

def getMTF(kspaceXY, z, ax, ay, num_aperture=0.004, useBessel=False):

    """
    Generates the MTF for the aberration state [z, ax, ay] with the given numerical aperture.
    Args:
        kspaceXY: 2D array of kx and ky wave vectors in a list of length 2 [kx, ky]
        z: defocus in um
        ax: amount of ax in um
        ay: amount of ay in um
        num_aperture: numerical aperture of the optical system
        useBessel: if yes, a bessel mtf will be returned, if no the gaussian approximation will be returned

    Returns:
        a 2D array of mtf with the zero frequency in the center.

    """

    kX, kY = kspaceXY

    qq = np.add(
        np.add(
            np.multiply(
                np.multiply(
                    kX, kX),
                z ** 2 + ax ** 2 + ay ** 2 + 2 * ax * z),
            np.multiply(
                np.multiply(
                    kY, kY),
                z ** 2 + ax ** 2 + ay ** 2 - 2 * ax * z)),
        np.multiply(
            np.multiply(
                kX, kY), - 4 * ay * z))

    if not useBessel:
        return np.exp(np.multiply
                      ((-1 / 8) * (num_aperture ** 2)
                       , qq)).astype(np.float64)
    else:
        ll = num_aperture * np.sqrt(np.add(qq, 10**-18))
        return np.divide(2 * special.jv(1, ll), ll).astype(np.float64)



import numpy as np
from . import mtfLib as mtfl
# import mtfLib as mtfl


def cost_mapfost(x, fftS, test_aberrs, numAper, stig_scale,useBessel, kSpaceXY):

    total_aberrs = [np.multiply(np.add(t, x), [1, stig_scale, stig_scale]) for t in test_aberrs]
    mtfS = [mtfl.getMTF(kSpaceXY, *qq, numAper, useBessel)
            for inx, qq in enumerate(total_aberrs)]
    m1 = np.array(mtfS[0])
    m2 = np.array(mtfS[1])
    f1 = fftS[0]
    f2 = fftS[1]
    log_lklhood = np.sum(np.abs(f1*m1 + f2*m2)** 2/(m1**2 + m2**2 + 0.000001))
    return -1 * log_lklhood




import numpy as np
import numpy.ma as ma
from scipy import ndimage

def low_pass_filter(pow_spec, crop_factor=0.25):
    """
    low pass filtering of a regularly edged power spectrum.

    Args:
        pow_spec: 2D array, fftshifted power spectrum.
        crop_factor:  ratio of the filtered edge length to the inital edge length.

    Returns: 2D array, filtered power spectrum

    """

    powSpEdgeLen = pow_spec.shape[1]
    patchEdge = int(crop_factor * powSpEdgeLen)
    pow_spec_cropped = pow_spec[int((powSpEdgeLen - patchEdge)/2):-1 * int((powSpEdgeLen - patchEdge)/2),
              int((powSpEdgeLen - patchEdge)/2):-1 * int((powSpEdgeLen - patchEdge)/2)]
    return pow_spec_cropped

def get_shift_vec(img_ref, img_corr):

    """
    estimates the shift vector between the two images
    Args:
        img_ref: 2D array of reference image
        img_corr:2d array of the second image

    Returns:
        returns a list of length 2 which is the shift vector from img_ref to img_corr.

    """

    
    img_ref_arr = np.asarray(img_ref)
    img_corr_arr = np.asarray(img_corr)

    img_ref_smoothed = ndimage.gaussian_filter(img_ref_arr, sigma=5)
    img_corr_smoothed = ndimage.gaussian_filter(img_corr_arr, sigma=5)

    img_ref_enh = img_ref_smoothed - img_ref_arr
    img_corr_enh = img_corr_smoothed - img_corr

    ft1 = np.fft.fft2(img_ref_enh)
    ft2 = np.fft.fft2(img_corr_enh)

    cross_corr = np.fft.fftshift(np.abs(np.fft.ifft2(np.multiply(ft1, np.conj(ft2)))))
    shift_xy = np.subtract(np.unravel_index(cross_corr.argmax(), cross_corr.shape),
                            [len(img_ref_enh)/2, len(img_ref_enh)/2])[::-1]

    return shift_xy

def getMaskS(edgeLen,inOutCross=[1,1,1], reverse=False):

    radius = int(edgeLen/2)
    maskS = []
    if inOutCross[0]:

        if reverse:
            inCircleBinaryMask = np.sign(np.abs(np.add(np.sign(np.subtract(radius ** 2,
                                                                            np.add(*np.square(
                                                                                np.meshgrid(range(-1 * radius, radius),
                                                                                            range(-1 * radius, radius)
                                                                                            )
                                                                                )))), -1))).astype(float)
        else:


            inCircleBinaryMask = np.sign(np.add(np.sign(np.subtract(radius ** 2,
                                              np.add(*np.square(np.meshgrid(range(-1 * radius, radius),
                                                                            range(-1 * radius, radius)
                                                                            )
                                                                )))),1)).astype(float)
        maskS.append(inCircleBinaryMask)

    if inOutCross[1]:

        if reverse:
            outCircleBinaryMask = np.sign(np.add(np.sign(np.subtract(radius ** 2,
                                              np.add(*np.square(np.meshgrid(range(-1 * radius, radius),
                                                                            range(-1 * radius, radius)
                                                                            )
                                                                )))),1)).astype(float)
        else:

            outCircleBinaryMask = np.sign(np.abs(np.add(np.sign(np.subtract(radius ** 2,
                                              np.add(*np.square(np.meshgrid(range(-1 * radius, radius),
                                                                            range(-1 * radius, radius)
                                                                            )
                                                                )))),-1))).astype(float)
        maskS.append(outCircleBinaryMask)
    if inOutCross[2]:

        if reverse:
            crossMask = np.abs(np.sign(np.abs(np.multiply(*np.meshgrid(range(-1 * radius, radius),
                                                                       range(-1 * radius, radius)
                                                                       )
                                                          )))).astype(float)
        else:
            crossMask = np.abs(np.subtract(np.sign(np.abs(np.multiply(*np.meshgrid(range(-1 * radius, radius),
                                                                       range(-1 * radius, radius)
                                                                       )
                                                          ))), 1)).astype(float)


        maskS.append(crossMask)

    return maskS


def applyMask(powSpec, maskS=None, inOutCross=[0,1,1], debug = False):
    if maskS is None:
        maskS = getMaskS(len(powSpec), inOutCross)
    if debug:
        for im, mask in enumerate(maskS):
            print(mask)
    for ii,mask in enumerate(maskS):
        powSpec = ma.masked_array(powSpec, mask)

    return powSpec

def removeMeanAndMeanSlope(im):

    imarr = np.array(im)
    deMeaned = np.subtract(imarr, np.mean(imarr))
    imwidth, imheight = deMeaned.shape
    deSloped = np.subtract(deMeaned,
                           np.add(*np.meshgrid(np.linspace(np.mean(deMeaned[0]), np.mean(deMeaned[-1]), imwidth),
                                               np.linspace(np.mean(deMeaned[:, 0]), np.mean(deMeaned[:, -1]),
                                                           imheight))))

    return deSloped